@@ -37,6 +37,7 @@
 #include <openssl/pem.h>
 
 #include <libloc/libloc.h>
+#include <libloc/address.h>
 #include <libloc/as.h>
 #include <libloc/as-list.h>
 #include <libloc/compat.h>
@@ -124,7 +125,10 @@ struct loc_database_enumerator {
 
 	// For subnet search and bogons
 	struct loc_network_list* stack;
-	struct loc_network* last_network;
+
+	// For bogons
+	struct in6_addr gap6_start;
+	struct in6_addr gap4_start;
 };
 
 static int loc_database_read_magic(struct loc_database* db) {
@@ -754,13 +758,8 @@ static int loc_database_fetch_network(struct loc_database* db, struct loc_networ
 			return -1;
 	}
 
-#ifdef ENABLE_DEBUG
-	if (r == 0) {
-		char* string = loc_network_str(*network);
-		DEBUG(db->ctx, "Got network %s\n", string);
-		free(string);
-	}
-#endif
+	if (r == 0)
+		DEBUG(db->ctx, "Got network %s\n", loc_network_str(*network));
 
 	return r;
 }
@@ -805,8 +804,8 @@ static int __loc_database_lookup(struct loc_database* db, const struct in6_addr*
 	off_t node_index;
 
 	// Follow the path
-	int bit = in6_addr_get_bit(address, level);
-	in6_addr_set_bit(network_address, level, bit);
+	int bit = loc_address_get_bit(address, level);
+	loc_address_set_bit(network_address, level, bit);
 
 	if (bit == 0)
 		node_index = be32toh(node->zero);
@@ -877,7 +876,7 @@ LOC_EXPORT int loc_database_lookup_from_string(struct loc_database* db,
 		const char* string, struct loc_network** network) {
 	struct in6_addr address;
 
-	int r = loc_parse_address(db->ctx, string, &address);
+	int r = loc_address_parse(&address, NULL, string);
 	if (r)
 		return r;
 
@@ -985,9 +984,6 @@ static void loc_database_enumerator_free(struct loc_database_enumerator* enumera
 	if (enumerator->stack)
 		loc_network_list_unref(enumerator->stack);
 
-	if (enumerator->last_network)
-		loc_network_unref(enumerator->last_network);
-
 	free(enumerator);
 }
 
@@ -1017,6 +1013,10 @@ LOC_EXPORT int loc_database_enumerator_new(struct loc_database_enumerator** enum
 		return r;
 	}
 
+	// Initialize bogon search
+	loc_address_reset(&e->gap6_start, AF_INET6);
+	loc_address_reset(&e->gap4_start, AF_INET);
+
 	DEBUG(e->ctx, "Database enumerator object allocated at %p\n", e);
 
 	*enumerator = e;
@@ -1168,6 +1168,10 @@ static int loc_database_enumerator_match_network(
 		return 0;
 	}
 
+	// Match if no filter criteria is configured
+	if (!enumerator->countries && !enumerator->asns && !enumerator->flags)
+		return 1;
+
 	// Check if the country code matches
 	if (enumerator->countries && !loc_country_list_empty(enumerator->countries)) {
 		const char* country_code = loc_network_get_country_code(network);
@@ -1234,7 +1238,7 @@ static int __loc_database_enumerator_next_network(
 		}
 
 		// Mark the bits on the path correctly
-		in6_addr_set_bit(&enumerator->network_address,
+		loc_address_set_bit(&enumerator->network_address,
 			(node->depth > 0) ? node->depth - 1 : 0, node->i);
 
 		DEBUG(enumerator->ctx, "Looking at node %jd\n", (intmax_t)node->offset);
@@ -1393,23 +1397,6 @@ static int __loc_database_enumerator_next_network_flattened(
 /*
 	This function finds all bogons (i.e. gaps) between the input networks
 */
-static int __loc_database_enumerator_find_bogons(struct loc_ctx* ctx,
-		struct loc_network* first, struct loc_network* second, struct loc_network_list* bogons) {
-	// We do not need to check if first < second because the graph is always ordered
-
-	// The last address of the first network + 1 is the start of the gap
-	struct in6_addr first_address = address_increment(loc_network_get_last_address(first));
-
-	// The first address of the second network - 1 is the end of the gap
-	struct in6_addr last_address = address_decrement(loc_network_get_first_address(second));
-
-	// If first address is now greater than last address, there is no gap
-	if (in6_addr_cmp(&first_address, &last_address) >= 0)
-		return 0;
-
-	return loc_network_list_summarize(ctx, &first_address, &last_address, &bogons);
-}
-
 static int __loc_database_enumerator_next_bogon(
 		struct loc_database_enumerator* enumerator, struct loc_network** bogon) {
 	int r;
@@ -1427,26 +1414,76 @@ static int __loc_database_enumerator_next_bogon(
 	}
 
 	struct loc_network* network = NULL;
+	struct in6_addr* gap_start = NULL;
+	struct in6_addr gap_end = IN6ADDR_ANY_INIT;
 
 	while (1) {
 		r = __loc_database_enumerator_next_network(enumerator, &network, 1);
 		if (r)
-			goto ERROR;
+			return r;
 
+		// We have read the last network
 		if (!network)
-			break;
+			goto FINISH;
 
-		if (enumerator->last_network && loc_network_address_family(enumerator->last_network) == loc_network_address_family(network)) {
-			r = __loc_database_enumerator_find_bogons(enumerator->ctx,
-				enumerator->last_network, network, enumerator->stack);
+		const char* country_code = loc_network_get_country_code(network);
+
+		/*
+			Skip anything that does not have a country code
+
+			Even if a network is part of the routing table, and the database provides
+			an ASN, this does not mean that this is a legitimate announcement.
+		*/
+		if (country_code && !*country_code) {
+			loc_network_unref(network);
+			continue;
+		}
+
+		// Determine the network family
+		int family = loc_network_address_family(network);
+
+		switch (family) {
+			case AF_INET6:
+				gap_start = &enumerator->gap6_start;
+				break;
+
+			case AF_INET:
+				gap_start = &enumerator->gap4_start;
+				break;
+
+			default:
+				ERROR(enumerator->ctx, "Unsupported network family %d\n", family);
+				errno = ENOTSUP;
+				return 1;
+		}
+
+		const struct in6_addr* first_address = loc_network_get_first_address(network);
+		const struct in6_addr* last_address = loc_network_get_last_address(network);
+
+		// Skip if this network is a subnet of a former one
+		if (loc_address_cmp(gap_start, last_address) >= 0) {
+			loc_network_unref(network);
+			continue;
+		}
+
+		// Search where the gap could end
+		gap_end = *first_address;
+		loc_address_decrement(&gap_end);
+
+		// There is a gap
+		if (loc_address_cmp(gap_start, &gap_end) <= 0) {
+			r = loc_network_list_summarize(enumerator->ctx,
+				gap_start, &gap_end, &enumerator->stack);
 			if (r) {
 				loc_network_unref(network);
-				goto ERROR;
+				return r;
 			}
 		}
 
-		// Remember network for next iteration
-		enumerator->last_network = loc_network_ref(network);
+		// The gap now starts after this network
+		*gap_start = *last_address;
+		loc_address_increment(gap_start);
+
 		loc_network_unref(network);
 
 		// Try to return something
@@ -1455,8 +1492,46 @@ static int __loc_database_enumerator_next_bogon(
 			break;
 	}
 
-ERROR:
-	return r;
+	return 0;
+
+FINISH:
+
+	if (!loc_address_all_zeroes(&enumerator->gap6_start)) {
+		r = loc_address_reset_last(&gap_end, AF_INET6);
+		if (r)
+			return r;
+
+		if (loc_address_cmp(&enumerator->gap6_start, &gap_end) <= 0) {
+			r = loc_network_list_summarize(enumerator->ctx,
+				&enumerator->gap6_start, &gap_end, &enumerator->stack);
+			if (r)
+				return r;
+		}
+
+		// Reset start
+		loc_address_reset(&enumerator->gap6_start, AF_INET6);
+	}
+
+	if (!loc_address_all_zeroes(&enumerator->gap4_start)) {
+		r = loc_address_reset_last(&gap_end, AF_INET);
+		if (r)
+			return r;
+
+		if (loc_address_cmp(&enumerator->gap4_start, &gap_end) <= 0) {
+			r = loc_network_list_summarize(enumerator->ctx,
+				&enumerator->gap4_start, &gap_end, &enumerator->stack);
+			if (r)
+				return r;
+		}
+
+		// Reset start
+		loc_address_reset(&enumerator->gap4_start, AF_INET);
+	}
+
+	// Try to return something
+	*bogon = loc_network_list_pop_first(enumerator->stack);
+
+	return 0;
 }
 
 LOC_EXPORT int loc_database_enumerator_next_network(
