@@ -26,6 +26,7 @@
 #endif
 
 #include <libloc/libloc.h>
+#include <libloc/address.h>
 #include <libloc/compat.h>
 #include <libloc/country.h>
 #include <libloc/network.h>
@@ -44,108 +45,40 @@ struct loc_network {
 	char country_code[3];
 	uint32_t asn;
 	enum loc_network_flags flags;
+
+	char string[INET6_ADDRSTRLEN + 4];
 };
 
-static int valid_prefix(struct in6_addr* address, unsigned int prefix) {
-	// The prefix cannot be larger than 128 bits
-	if (prefix > 128)
-		return 1;
-
-	// And the prefix cannot be zero
-	if (prefix == 0)
-		return 1;
-
-	// For IPv4-mapped addresses the prefix has to be 96 or lager
-	if (IN6_IS_ADDR_V4MAPPED(address) && prefix <= 96)
-		return 1;
-
-	return 0;
-}
-
-static struct in6_addr prefix_to_bitmask(unsigned int prefix) {
-	struct in6_addr bitmask;
-
-	for (unsigned int i = 0; i < 16; i++)
-		bitmask.s6_addr[i] = 0;
-
-	for (int i = prefix, j = 0; i > 0; i -= 8, j++) {
-		if (i >= 8)
-			bitmask.s6_addr[j] = 0xff;
-		else
-			bitmask.s6_addr[j] = 0xff << (8 - i);
-	}
-
-	return bitmask;
-}
-
-static struct in6_addr make_first_address(const struct in6_addr* address, const struct in6_addr* bitmask) {
-	struct in6_addr a;
-
-	// Perform bitwise AND
-	for (unsigned int i = 0; i < 4; i++)
-		a.s6_addr32[i] = address->s6_addr32[i] & bitmask->s6_addr32[i];
-
-	return a;
-}
-
-static struct in6_addr make_last_address(const struct in6_addr* address, const struct in6_addr* bitmask) {
-	struct in6_addr a;
-
-	// Perform bitwise OR
-	for (unsigned int i = 0; i < 4; i++)
-		a.s6_addr32[i] = address->s6_addr32[i] | ~bitmask->s6_addr32[i];
-
-	return a;
-}
-
 LOC_EXPORT int loc_network_new(struct loc_ctx* ctx, struct loc_network** network,
 		struct in6_addr* address, unsigned int prefix) {
-	// Address cannot be unspecified
-	if (IN6_IS_ADDR_UNSPECIFIED(address)) {
-		DEBUG(ctx, "Start address is unspecified\n");
-		return -EINVAL;
-	}
-
-	// Address cannot be loopback
-	if (IN6_IS_ADDR_LOOPBACK(address)) {
-		DEBUG(ctx, "Start address is loopback address\n");
-		return -EINVAL;
-	}
-
-	// Address cannot be link-local
-	if (IN6_IS_ADDR_LINKLOCAL(address)) {
-		DEBUG(ctx, "Start address cannot be link-local\n");
-		return -EINVAL;
-	}
-
-	// Address cannot be site-local
-	if (IN6_IS_ADDR_SITELOCAL(address)) {
-		DEBUG(ctx, "Start address cannot be site-local\n");
-		return -EINVAL;
-	}
-
 	// Validate the prefix
-	if (valid_prefix(address, prefix) != 0) {
-		DEBUG(ctx, "Invalid prefix: %u\n", prefix);
-		return -EINVAL;
+	if (!loc_address_valid_prefix(address, prefix)) {
+		ERROR(ctx, "Invalid prefix in %s: %u\n", loc_address_str(address), prefix);
+		errno = EINVAL;
+		return 1;
 	}
 
 	struct loc_network* n = calloc(1, sizeof(*n));
-	if (!n)
-		return -ENOMEM;
+	if (!n) {
+		errno = ENOMEM;
+		return 1;
+	}
 
 	n->ctx = loc_ref(ctx);
 	n->refcount = 1;
 
 	// Store the prefix
-	n->prefix = prefix;
+	if (IN6_IS_ADDR_V4MAPPED(address))
+		n->prefix = prefix + 96;
+	else
+		n->prefix = prefix;
 
 	// Convert the prefix into a bitmask
-	struct in6_addr bitmask = prefix_to_bitmask(n->prefix);
+	struct in6_addr bitmask = loc_prefix_to_bitmask(n->prefix);
 
 	// Store the first and last address in the network
-	n->first_address = make_first_address(address, &bitmask);
-	n->last_address = make_last_address(&n->first_address, &bitmask);
+	n->first_address = loc_address_and(address, &bitmask);
+	n->last_address  = loc_address_or(&n->first_address, &bitmask);
 
 	// Set family
 	n->family = loc_address_family(&n->first_address);
@@ -155,56 +88,20 @@ LOC_EXPORT int loc_network_new(struct loc_ctx* ctx, struct loc_network** network
 	return 0;
 }
 
-LOC_EXPORT int loc_network_new_from_string(struct loc_ctx* ctx, struct loc_network** network,
-		const char* address_string) {
-	struct in6_addr first_address;
-	char* prefix_string;
-	unsigned int prefix = 128;
-	int r = -EINVAL;
+LOC_EXPORT int loc_network_new_from_string(struct loc_ctx* ctx,
+		struct loc_network** network, const char* string) {
+	struct in6_addr address;
+	unsigned int prefix;
 
-	DEBUG(ctx, "Attempting to parse network %s\n", address_string);
-
-	// Make a copy of the string to work on it
-	char* buffer = strdup(address_string);
-	address_string = prefix_string = buffer;
-
-	// Split address and prefix
-	address_string = strsep(&prefix_string, "/");
-
-	DEBUG(ctx, "  Split into address = %s, prefix = %s\n", address_string, prefix_string);
-
-	// Parse the address
-	r = loc_parse_address(ctx, address_string, &first_address);
+	// Parse the input
+	int r = loc_address_parse(&address, &prefix, string);
 	if (r) {
-		DEBUG(ctx, "The address could not be parsed\n");
-		goto FAIL;
-	}
-
-	// If a prefix was given, we will try to parse it
-	if (prefix_string) {
-		// Convert prefix to integer
-		prefix = strtol(prefix_string, NULL, 10);
-
-		if (!prefix) {
-			DEBUG(ctx, "The prefix was not parsable: %s\n", prefix_string);
-			goto FAIL;
-		}
-
-		// Map the prefix to IPv6 if needed
-		if (IN6_IS_ADDR_V4MAPPED(&first_address))
-			prefix += 96;
-	}
-
-FAIL:
-	// Free temporary buffer
-	free(buffer);
-
-	// Exit if the parsing was unsuccessful
-	if (r)
+		ERROR(ctx, "Could not parse network %s: %m\n", string);
 		return r;
+	}
 
 	// Create a new network
-	return loc_network_new(ctx, network, &first_address, prefix);
+	return loc_network_new(ctx, network, &address, prefix);
 }
 
 LOC_EXPORT struct loc_network* loc_network_ref(struct loc_network* network) {
@@ -231,61 +128,27 @@ LOC_EXPORT struct loc_network* loc_network_unref(struct loc_network* network) {
 	return NULL;
 }
 
-static int format_ipv6_address(const struct in6_addr* address, char* string, size_t length) {
-	const char* ret = inet_ntop(AF_INET6, address, string, length);
-	if (!ret)
-		return -1;
+LOC_EXPORT const char* loc_network_str(struct loc_network* network) {
+	if (!*network->string) {
+		// Format the address
+		const char* address = loc_address_str(&network->first_address);
+		if (!address)
+			return NULL;
 
-	return 0;
-}
+		// Fetch the prefix
+		unsigned int prefix = loc_network_prefix(network);
 
-static int format_ipv4_address(const struct in6_addr* address, char* string, size_t length) {
-	struct in_addr ipv4_address;
-	ipv4_address.s_addr = address->s6_addr32[3];
-
-	const char* ret = inet_ntop(AF_INET, &ipv4_address, string, length);
-	if (!ret)
-		return -1;
-
-	return 0;
-}
-
-LOC_EXPORT char* loc_network_str(struct loc_network* network) {
-	int r;
-	const size_t length = INET6_ADDRSTRLEN + 4;
-
-	char* string = malloc(length);
-	if (!string)
-		return NULL;
-
-	unsigned int prefix = network->prefix;
-
-	switch (network->family) {
-		case AF_INET6:
-			r = format_ipv6_address(&network->first_address, string, length);
-			break;
-
-		case AF_INET:
-			r = format_ipv4_address(&network->first_address, string, length);
-			prefix -= 96;
-			break;
-
-		default:
-			r = -1;
-			break;
+		// Format the string
+		int r = snprintf(network->string, sizeof(network->string) - 1,
+			"%s/%u", address, prefix);
+		if (r < 0) {
+			ERROR(network->ctx, "Could not format network string: %m\n");
+			*network->string = '\0';
+			return NULL;
+		}
 	}
 
-	if (r) {
-		ERROR(network->ctx, "Could not convert network to string: %s\n", strerror(errno));
-		free(string);
-
-		return NULL;
-	}
-
-	// Append prefix
-	sprintf(string + strlen(string), "/%u", prefix);
-
-	return string;
+	return network->string;
 }
 
 LOC_EXPORT int loc_network_address_family(struct loc_network* network) {
@@ -304,62 +167,29 @@ LOC_EXPORT unsigned int loc_network_prefix(struct loc_network* network) {
 	return 0;
 }
 
-static char* loc_network_format_address(struct loc_network* network, const struct in6_addr* address) {
-	const size_t length = INET6_ADDRSTRLEN;
-
-	char* string = malloc(length);
-	if (!string)
-		return NULL;
-
-	int r = 0;
-
-	switch (network->family) {
-		case AF_INET6:
-			r = format_ipv6_address(address, string, length);
-			break;
-
-		case AF_INET:
-			r = format_ipv4_address(address, string, length);
-			break;
-
-		default:
-			r = -1;
-			break;
-	}
-
-	if (r) {
-		ERROR(network->ctx, "Could not format IP address to string: %s\n", strerror(errno));
-		free(string);
-
-		return NULL;
-	}
-
-	return string;
-}
-
 LOC_EXPORT const struct in6_addr* loc_network_get_first_address(struct loc_network* network) {
 	return &network->first_address;
 }
 
-LOC_EXPORT char* loc_network_format_first_address(struct loc_network* network) {
-	return loc_network_format_address(network, &network->first_address);
+LOC_EXPORT const char* loc_network_format_first_address(struct loc_network* network) {
+	return loc_address_str(&network->first_address);
 }
 
 LOC_EXPORT const struct in6_addr* loc_network_get_last_address(struct loc_network* network) {
 	return &network->last_address;
 }
 
-LOC_EXPORT char* loc_network_format_last_address(struct loc_network* network) {
-	return loc_network_format_address(network, &network->last_address);
+LOC_EXPORT const char* loc_network_format_last_address(struct loc_network* network) {
+	return loc_address_str(&network->last_address);
 }
 
 LOC_EXPORT int loc_network_matches_address(struct loc_network* network, const struct in6_addr* address) {
 	// Address must be larger than the start address
-	if (in6_addr_cmp(&network->first_address, address) > 0)
+	if (loc_address_cmp(&network->first_address, address) > 0)
 		return 0;
 
 	// Address must be smaller than the last address
-	if (in6_addr_cmp(&network->last_address, address) < 0)
+	if (loc_address_cmp(&network->last_address, address) < 0)
 		return 0;
 
 	// The address is inside this network
@@ -425,7 +255,7 @@ LOC_EXPORT int loc_network_set_flag(struct loc_network* network, uint32_t flag)
 
 LOC_EXPORT int loc_network_cmp(struct loc_network* self, struct loc_network* other) {
 	// Compare address
-	int r = in6_addr_cmp(&self->first_address, &other->first_address);
+	int r = loc_address_cmp(&self->first_address, &other->first_address);
 	if (r)
 		return r;
 
@@ -464,12 +294,12 @@ LOC_EXPORT int loc_network_is_subnet(struct loc_network* self, struct loc_networ
 
 	// If the start address of the other network is smaller than this network,
 	// it cannot be a subnet.
-	if (in6_addr_cmp(&self->first_address, &other->first_address) > 0)
+	if (loc_address_cmp(&self->first_address, &other->first_address) > 0)
 		return 0;
 
 	// If the end address of the other network is greater than this network,
 	// it cannot be a subnet.
-	if (in6_addr_cmp(&self->last_address, &other->last_address) < 0)
+	if (loc_address_cmp(&self->last_address, &other->last_address) < 0)
 		return 0;
 
 	return 1;
@@ -485,7 +315,7 @@ LOC_EXPORT int loc_network_subnets(struct loc_network* network,
 	unsigned int prefix = network->prefix + 1;
 
 	// Check if the new prefix is valid
-	if (valid_prefix(&network->first_address, prefix))
+	if (!loc_address_valid_prefix(&network->first_address, prefix))
 		return -1;
 
 	// Create the first half of the network
@@ -494,7 +324,8 @@ LOC_EXPORT int loc_network_subnets(struct loc_network* network,
 		return r;
 
 	// The next subnet starts after the first one
-	struct in6_addr first_address = address_increment(&(*subnet1)->last_address);
+	struct in6_addr first_address = (*subnet1)->last_address;
+	loc_address_increment(&first_address);
 
 	// Create the second half of the network
 	r = loc_network_new(network->ctx, subnet2, &first_address, prefix);
@@ -600,15 +431,8 @@ LOC_EXPORT struct loc_network_list* loc_network_exclude(
 		struct loc_network* self, struct loc_network* other) {
 	struct loc_network_list* list;
 
-#ifdef ENABLE_DEBUG
-	char* n1 = loc_network_str(self);
-	char* n2 = loc_network_str(other);
-
-	DEBUG(self->ctx, "Returning %s excluding %s...\n", n1, n2);
-
-	free(n1);
-	free(n2);
-#endif
+	DEBUG(self->ctx, "Returning %s excluding %s...\n",
+		loc_network_str(self), loc_network_str(other));
 
 	// Create a new list with the result
 	int r = loc_network_list_new(self->ctx, &list);
@@ -744,6 +568,10 @@ int loc_network_new_from_database_v1(struct loc_ctx* ctx, struct loc_network** n
 		struct in6_addr* address, unsigned int prefix, const struct loc_database_network_v1* dbobj) {
 	char country_code[3] = "\0\0";
 
+	// Adjust prefix for IPv4
+	if (IN6_IS_ADDR_V4MAPPED(address))
+		prefix -= 96;
+
 	int r = loc_network_new(ctx, network, address, prefix);
 	if (r) {
 		ERROR(ctx, "Could not allocate a new network: %s", strerror(-r));
@@ -842,7 +670,7 @@ static struct loc_network_tree_node* loc_network_tree_get_path(struct loc_networ
 
 	for (unsigned int i = 0; i < prefix; i++) {
 		// Check if the ith bit is one or zero
-		node = loc_network_tree_get_node(node, in6_addr_get_bit(address, i));
+		node = loc_network_tree_get_node(node, loc_address_get_bit(address, i));
 	}
 
 	return node;
@@ -913,12 +741,11 @@ struct loc_network_tree* loc_network_tree_unref(struct loc_network_tree* tree) {
 static int __loc_network_tree_dump(struct loc_network* network, void* data) {
 	DEBUG(network->ctx, "Dumping network at %p\n", network);
 
-	char* s = loc_network_str(network);
+	const char* s = loc_network_str(network);
 	if (!s)
 		return 1;
 
 	INFO(network->ctx, "%s\n", s);
-	free(s);
 
 	return 0;
 }
